# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""l3p ip pool size

Revision ID: 27b724002081
Revises: cb5b16acbeb0
Create Date: 2017-07-06 17:34:18.856803

"""

# revision identifiers, used by Alembic.
revision = '27b724002081'
down_revision = 'cb5b16acbeb0'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('gp_l3_policies', 'ip_pool',
                    existing_type=sa.String(length=64),
                    type_=sa.String(length=255))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa


class ApicIntraPtgDB(model_base.BASEV2):
    __tablename__ = 'gp_apic_intra_ptg'
    policy_target_group_id = sa.Column(
        sa.String(36), sa.ForeignKey('gp_policy_target_groups.id',
                                     ondelete='CASCADE'), primary_key=True)
    intra_ptg_allow = sa.Column(sa.Boolean, default=True, nullable=False)


class ApicIntraPtgDBMixin(object):

    def get_intra_ptg_allow(self, session, policy_target_group_id):
        row = (session.query(ApicIntraPtgDB).filter_by(
               policy_target_group_id=policy_target_group_id).one())
        return row['intra_ptg_allow']

    def set_intra_ptg_allow(self, session, policy_target_group_id,
                            intra_ptg_allow=True):
        with session.begin(subtransactions=True):
            row = (session.query(ApicIntraPtgDB).filter_by(
                policy_target_group_id=policy_target_group_id).first())
            if not row:
                row = ApicIntraPtgDB(
                    policy_target_group_id=policy_target_group_id,
                    intra_ptg_allow=intra_ptg_allow)
                session.add(row)
            else:
                row.update({'intra_ptg_allow': intra_ptg_allow})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg
import unittest2

from gbpclient.v2_0 import client as gbp_client
from gbpservice.nfp.orchestrator.openstack import openstack_driver
from keystoneauth1.identity import v2
from keystoneauth1 import session
from keystoneclient.v2_0 import client as identity_client
from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client

cfg.CONF.import_group('nfp_keystone_authtoken',
                      'gbpservice.nfp.orchestrator.modules.__init__')


class SampleData(unittest2.TestCase):

    def __init__(self, *args, **kwargs):
        super(SampleData, self).__init__(*args, **kwargs)
        self.AUTH_TOKEN = '6db9dfa4d29d442eb2b23811ad4b3a6d'
        self.AUTH_URL = 'http://localhost:5000/v2.0/'
        self.ENDPOINT_URL = 'http://localhost:9696/'
        self.FLAVOR_NAME = 'm1.tiny'
        self.IMAGE_NAME = 'cirros-0.3.4-x86_64-uec'
        self.IMAGE_ID = '7022c5a4-ef0c-4f7e-a2c8-b7f5b36c9086'
        self.INSTANCE_ID = '60c7ebc4-aa70-4ee6-aad6-41e99d27ceec'
        self.PASSWORD = 'admin_pass'
        self.PORT_ID = '16fa0e95-3c7a-4dd6-87bd-c76b14f9eac2'
        self.TENANT_ID = '384757095ca4495683c7f34ae077f8c0'
        self.TENANT_NAME = 'admin'
        self.USERNAME = 'admin'


@mock.patch.object(identity_client, "Client")
class TestKeystoneClient(SampleData):

    def __init__(self, *args, **kwargs):
        super(TestKeystoneClient, self).__init__(*args, **kwargs)
        self.keystone_obj = openstack_driver.KeystoneClient(cfg.CONF)

    def setUp(self):
        cfg.CONF.set_override('admin_user',
                              'neutron',
                              group='nfp_keystone_authtoken')
        cfg.CONF.set_override('admin_password',
                              'neutron_pass',
                              group='nfp_keystone_authtoken')
        cfg.CONF.set_override('admin_tenant_name',
                              'service',
                              group='nfp_keystone_authtoken')
        cfg.CONF.set_override('auth_version',
                              'None',
                              group='nfp_keystone_authtoken')

    @mock.patch.object(v2, "Password")
    @mock.patch.object(session.Session, "get_token")
    def test_get_admin_token(self, mock_session, mock_v2, mock_obj):
        mock_session.return_value = True
        retval = self.keystone_obj.get_admin_token()
        self.assertTrue(retval)
        mock_v2.assert_called_once_with(auth_url=self.AUTH_URL,
                                        password='neutron_pass',
                                        tenant_name='service',
                                        username='neutron')

    @mock.patch.object(v2, "Password")
    @mock.patch.object(session.Session, "get_token")
    def test_get_scoped_keystone_token(self, mock_session, mock_v2, mock_obj):
        mock_session.return_value = True
        retval = self.keystone_obj.get_scoped_keystone_token(self.USERNAME,
                                                             self.PASSWORD,
                                                             self.TENANT_NAME,
                                                             self.TENANT_ID)
        self.assertTrue(retval)
        mock_v2.assert_called_once_with(auth_url=self.AUTH_URL,
                                        password=self.PASSWORD,
                                        tenant_name=self.TENANT_NAME,
                                        username=self.USERNAME)

    @mock.patch.object(v2, "Password")
    @mock.patch.object(session.Session, "get_token")
    def test_get_tenant_id(self, mock_session, mock_v2, mock_obj):
        instance = mock_obj.return_value
        instance.tenants.find().id = True
        retval = self.keystone_obj.get_tenant_id(self.AUTH_TOKEN,
                                                 "service")
        self.assertTrue(retval)
        mock_obj.assert_called_once_with(session=mock.ANY)


@mock.patch.object(nova_client, "Client")
class TestNovaClient(SampleData):

    def __init__(self, *args, **kwargs):
        super(TestNovaClient, self).__init__(*args, **kwargs)
        self.nova_obj = openstack_driver.NovaClient(cfg.CONF)

    @mock.patch.object(identity_client, "Client")
    def test_get_image_id(self, key_obj, mock_obj):
        instance = mock_obj.return_value
        key_obj.return_value = True
        instance.images.find().id = True
        retval = self.nova_obj.get_image_id(self.AUTH_TOKEN,
                                            self.TENANT_ID,
                                            self.IMAGE_NAME)
        self.assertTrue(retval)
        mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                         tenant_id=self.TENANT_ID,
                                         auth_url=self.AUTH_URL)

    def test_get_flavor_id(self, mock_obj):
        instance = mock_obj.return_value
        instance.flavors.find().id = True
        retval = self.nova_obj.get_flavor_id(self.AUTH_TOKEN,
                                             self.TENANT_ID,
                                             self.FLAVOR_NAME)
        self.assertTrue(retval)

    @mock.patch.object(identity_client, "Client")
    def test_get_instance(self, key_obj, mock_obj):
        instance = mock_obj.return_value
        obj = instance.servers.get(self.INSTANCE_ID).to_dict()
        key_obj.return_value = self.AUTH_TOKEN
        retval = self.nova_obj.get_instance(self.AUTH_TOKEN,
                                            self.TENANT_ID,
                                            self.INSTANCE_ID)
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                         tenant_id=self.TENANT_ID,
                                         auth_url=self.AUTH_URL)

    def test_get_keypair(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.keypairs.find(name="keypair_name").to_dict()
        retval = self.nova_obj.get_keypair(self.AUTH_TOKEN,
                                           self.TENANT_ID,
                                           "keypair_name")
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                         tenant_id=self.TENANT_ID,
                                         auth_url=self.AUTH_URL)

    def test_attach_interface(self, mock_obj):
        instance = mock_obj.return_value
        with mock.patch.object(instance.servers,
                               "interface_attach") as mock_obj1:
            mock_obj1.return_value = True
            retval = self.nova_obj.attach_interface(self.AUTH_TOKEN,
                                                    self.TENANT_ID,
                                                    self.INSTANCE_ID,
                                                    "port_id")
            self.assertTrue(retval)
            mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                             tenant_id=self.TENANT_ID,
                                             auth_url=self.AUTH_URL)

    def test_detach_interface(self, mock_obj):
        instance = mock_obj.return_value
        with mock.patch.object(instance.servers,
                               "interface_detach") as mock_obj1:
            mock_obj1.return_value = True
            retval = self.nova_obj.detach_interface(self.AUTH_TOKEN,
                                                    self.TENANT_ID,
                                                    self.INSTANCE_ID,
                                                    "port_id")

            self.assertTrue(retval)
            mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                             tenant_id=self.TENANT_ID,
                                             auth_url=self.AUTH_URL)

    def test_delete_instance(self, mock_obj):
        instance = mock_obj.return_value
        with mock.patch.object(instance.servers, "delete") as mock_obj1:
            mock_obj1.return_value = True
            retval = self.nova_obj.delete_instance(self.AUTH_TOKEN,
                                                   self.TENANT_ID,
                                                   self.INSTANCE_ID)

            self.assertIsNone(retval)
            mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                             tenant_id=self.TENANT_ID,
                                             auth_url=self.AUTH_URL)

    def test_get_instances(self, mock_obj):
        instance = mock_obj.return_value
        instance.servers.list("instance_test").to_dict()
        retval = self.nova_obj.get_instances(self.AUTH_TOKEN,
                                             {'tenant_id': self.TENANT_ID})
        self.assertIsNotNone(retval)
        mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                         tenant_id=self.TENANT_ID,
                                         auth_url=self.AUTH_URL)

    def test_create_instance(self, mock_obj):
        instance = mock_obj.return_value
        instance.flavors.find(self.FLAVOR_NAME)
        with mock.patch.object(instance.servers, "create") as mock_obj1:
            instance = mock_obj1.return_value
            obj1 = instance.to_dict()['id']
            retval = self.nova_obj.create_instance(self.AUTH_TOKEN,
                                                   self.TENANT_ID,
                                                   self.IMAGE_ID,
                                                   self.FLAVOR_NAME,
                                                   None,
                                                   "name",
                                                   False,
                                                   '2',
                                                   None,
                                                   None,
                                                   None,
                                                   False,
                                                   None,
                                                   '',
                                                   None
                                                   )

            self.assertEqual(retval, obj1)
            mock_obj.assert_called_once_with('2', auth_token=self.AUTH_TOKEN,
                                             tenant_id=self.TENANT_ID,
                                             auth_url=self.AUTH_URL)


@mock.patch.object(neutron_client, "Client")
class TestNeutronClient(SampleData):

    def __init__(self, *args, **kwargs):
        super(TestNeutronClient, self).__init__(*args, **kwargs)
        self.neutron_obj = openstack_driver.NeutronClient(cfg.CONF)

    def test_get_floating_ip(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_floatingip('floatingip_id')['floatingip']
        retval = self.neutron_obj.get_floating_ip(
            self.AUTH_TOKEN, 'floatingip_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_floating_ips(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_floatingips()['floatingips']
        filters = {'tenant_id': self.TENANT_ID,
                   'port_id': self.PORT_ID}
        retval = self.neutron_obj.get_floating_ips(self.AUTH_TOKEN,
                                                   **filters)
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_port(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_port(self.PORT_ID)
        retval = self.neutron_obj.get_port(self.AUTH_TOKEN,
                                           self.PORT_ID)
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_ports(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_ports().get('ports', [])
        retval = self.neutron_obj.get_ports(self.AUTH_TOKEN,
                                            {})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_subnets(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_subnets().get('subnets', [])
        retval = self.neutron_obj.get_subnets(self.AUTH_TOKEN, {})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_pools(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_pools().get('pools', [])
        retval = self.neutron_obj.get_pools(self.AUTH_TOKEN,
                                            {})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_vip(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_vip('vip_id')
        retval = self.neutron_obj.get_vip(self.AUTH_TOKEN,
                                          'vip_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_subnet(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_subnet('subnet_id')
        retval = self.neutron_obj.get_subnet(self.AUTH_TOKEN,
                                             'subnet_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_delete_floatingip(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.delete_floatingip('floatingip_id')
        retval = self.neutron_obj.delete_floatingip(self.AUTH_TOKEN,
                                                    'floatingip_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_update_port(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.update_port('port_id', body='port_info')
        retval = self.neutron_obj.update_port(self.AUTH_TOKEN,
                                              'port_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_floating_ips_for_ports(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_floatingips('port_id')
        retval = self.neutron_obj.get_floating_ips_for_ports(
            self.AUTH_TOKEN)
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_disassociate_floating_ip(self, mock_obj):
        instance = mock_obj.return_value
        instance.update_floatingip('floatingip_id', body='data')
        retval = self.neutron_obj.disassociate_floating_ip(self.AUTH_TOKEN,
                                                           'floatingip_id')
        self.assertIsNone(retval)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_associate_floating_ip(self, mock_obj):
        instance = mock_obj.return_value
        instance.update_floatingip('floatingip_id', body='data')
        retval = self.neutron_obj.associate_floating_ip(self.AUTH_TOKEN,
                                                        'floatingip_id',
                                                        'port_id')
        self.assertIsNone(retval)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_list_ports(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_ports(id='port_ids').get('ports', [])
        retval = self.neutron_obj.list_ports(self.AUTH_TOKEN,
                                             'port_ids=[]')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_list_subnets(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_subnets(id='subnet_ids').get('subnets', [])
        retval = self.neutron_obj.list_subnets(self.AUTH_TOKEN,
                                               'subnet_ids=[]')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_create_port(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.create_port(body='attr')['port']
        retval = self.neutron_obj.create_port(self.AUTH_TOKEN,
                                              self.TENANT_ID,
                                              'net_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_delete_port(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.delete_port('port_id')
        retval = self.neutron_obj.delete_port(self.AUTH_TOKEN,
                                              'port_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)


@mock.patch.object(gbp_client, "Client")
class TestGBPClient(SampleData):

    def __init__(self, *args, **kwargs):
        super(TestGBPClient, self).__init__(*args, **kwargs)
        self.gbp_obj = openstack_driver.GBPClient(cfg.CONF)

    def test_get_policy_target_groups(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_policy_target_groups()['policy_target_groups']
        retval = self.gbp_obj.get_policy_target_groups(self.AUTH_TOKEN, {})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_policy_target_group(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_policy_target_group(
            'ptg_id', {})['policy_target_group']
        retval = self.gbp_obj.get_policy_target_group(self.AUTH_TOKEN,
                                                      'ptg_id', {})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_update_policy_target_group(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.update_policy_target_group(
            body='policy_target_group_info')['policy_target_group']
        retval = self.gbp_obj.update_policy_target_group(
            self.AUTH_TOKEN,
            'ptg_id',
            'policy_target_group_info')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_update_policy_target(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.update_policy_target('policy_target_id',
                                            body='policy_target_info')[
            'policy_target']
        retval = self.gbp_obj.update_policy_target(self.AUTH_TOKEN,
                                                   'policy_target_id',
                                                   'updated_pt')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_create_policy_target(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.create_policy_target(
            body='policy_target_info')['policy_target']
        retval = self.gbp_obj.create_policy_target(self.AUTH_TOKEN,
                                                   self.TENANT_ID,
                                                   'policy_target_group_id',
                                                   'name', port_id=None)
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_create_policy_target_group(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.create_policy_target_group(
            body='policy_target_group_info')['policy_target_group']
        retval = self.gbp_obj.create_policy_target_group(
            self.AUTH_TOKEN,
            self.TENANT_ID,
            'name',
            l2_policy_id=None)
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_delete_policy_target(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.delete_policy_target('policy_target_id')
        retval = self.gbp_obj.delete_policy_target(self.AUTH_TOKEN,
                                                   'policy_target_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_delete_policy_target_group(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.delete_policy_target_group('policy_target_id')
        retval = self.gbp_obj.delete_policy_target_group(
            self.AUTH_TOKEN,
            'policy_target_group_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_create_l2_policy(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.create_l2_policy(body='l2_policy_info')['l2_policy']
        retval = self.gbp_obj.create_l2_policy(self.AUTH_TOKEN,
                                               self.TENANT_ID,
                                               'name',
                                               l3_policy_id=None)
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_delete_l2_policy(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.delete_l2_policy('l2_policy_id')
        retval = self.gbp_obj.delete_l2_policy(self.AUTH_TOKEN,
                                               'l2_policy_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_l2_policys(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_l2_policies({})['l2_policies']
        retval = self.gbp_obj.get_l2_policys(self.AUTH_TOKEN,
                                             filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_l2_policy(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_l2_policy(
            'policy_id', {})['l2_policy']
        retval = self.gbp_obj.get_l2_policy(self.AUTH_TOKEN,
                                            'policy_id',
                                            filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_create_network_service_policy(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.create_network_service_policy(
            body='network_service_policy_info')['network_service_policy']
        retval = self.gbp_obj.create_network_service_policy(
            self.AUTH_TOKEN,
            'network_service_policy_info')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_network_service_policies(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_network_service_policies({})[
            'network_service_policies']
        retval = self.gbp_obj.get_network_service_policies(self.AUTH_TOKEN,
                                                           filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_external_policies(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_external_policies({})['external_policies']
        retval = self.gbp_obj.get_external_policies(self.AUTH_TOKEN,
                                                    filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_policy_rule_sets(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_policy_rule_sets({})['policy_rule_sets']
        retval = self.gbp_obj.get_policy_rule_sets(self.AUTH_TOKEN,
                                                   filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_policy_actions(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_policy_actions({})['policy_actions']
        retval = self.gbp_obj.get_policy_actions(self.AUTH_TOKEN,
                                                 filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_policy_rules(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_policy_rules({})['policy_rules']
        retval = self.gbp_obj.get_policy_rules(self.AUTH_TOKEN,
                                               filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_create_l3_policy(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.create_l3_policy(
            body='l3_policy_info')['l3_policy']
        retval = self.gbp_obj.create_l3_policy(self.AUTH_TOKEN,
                                               'l3_policy_info')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_l3_policy(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_l3_policy('policy_id',
                                      {})['l3_policy']
        retval = self.gbp_obj.get_l3_policy(self.AUTH_TOKEN,
                                            'policy_id',
                                            filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_l3_policies(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_l3_policies({})['l3_policy']
        retval = self.gbp_obj.get_l3_policies(self.AUTH_TOKEN,
                                              filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_policy_targets(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_policy_targets({})['policy_targets']
        retval = self.gbp_obj.get_policy_targets(self.AUTH_TOKEN,
                                                 filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_list_pt(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.list_policy_targets({})['policy_targets']
        retval = self.gbp_obj.list_pt(self.AUTH_TOKEN,
                                      filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_policy_target(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_policy_target('pt_id',
                                          {})['policy_target']
        retval = self.gbp_obj.get_policy_target(self.AUTH_TOKEN,
                                                'pt_id',
                                                filters={})
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_service_profile(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_service_profile(
            'service_profile_id')['service_profile']
        retval = self.gbp_obj.get_service_profile(self.AUTH_TOKEN,
                                                  'service_profile_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_servicechain_node(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_servicechain_node(
            'node_id')['servicechain_node']
        retval = self.gbp_obj.get_servicechain_node(self.AUTH_TOKEN,
                                                    'node_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)

    def test_get_servicechain_instance(self, mock_obj):
        instance = mock_obj.return_value
        obj = instance.show_servicechain_instance(
            'instance_id')['servicechain_instance']
        retval = self.gbp_obj.get_servicechain_instance(self.AUTH_TOKEN,
                                                        'instance_id')
        self.assertEqual(retval, obj)
        mock_obj.assert_called_once_with(token=self.AUTH_TOKEN,
                                         endpoint_url=self.ENDPOINT_URL)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
import time
import warnings

# Override unnecessary Octavia config import
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v2.haproxy \
    import config
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v2.haproxy.\
    config import cfg
sys.modules['octavia.common.config'] = config
sys.modules['octavia.common.config.cfg'] = cfg

from octavia.amphorae.driver_exceptions import exceptions as driver_except
from octavia.amphorae.drivers.haproxy import rest_api_driver
from octavia.common.jinja.haproxy import jinja_cfg
from octavia.i18n import _LE
from octavia.i18n import _LW
from oslo_config import cfg
import requests

from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v2.haproxy.\
    local_cert_manager import LocalCertManager
from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)
API_VERSION = rest_api_driver.API_VERSION
OCTAVIA_API_CLIENT = rest_api_driver.OCTAVIA_API_CLIENT

CONF = cfg.CONF
CONF.import_group('haproxy_amphora', 'octavia.common.config')


class HaproxyAmphoraLoadBalancerDriver(
        rest_api_driver.HaproxyAmphoraLoadBalancerDriver):

    def __init__(self):
        super(rest_api_driver.HaproxyAmphoraLoadBalancerDriver,
              self).__init__()
        self.client = AmphoraAPIClient()
        self.cert_manager = LocalCertManager()
        self.jinja = jinja_cfg.JinjaTemplater(
            base_amp_path=CONF.haproxy_amphora.base_path,
            base_crt_dir=CONF.haproxy_amphora.base_cert_dir,
            haproxy_template=CONF.haproxy_amphora.haproxy_template)


class AmphoraAPIClient(rest_api_driver.AmphoraAPIClient):
    """Removed SSL verification from original api client"""

    def __init__(self):
        super(AmphoraAPIClient, self).__init__()
        self.session = requests.Session()

    def _base_url(self, ip):
        return "http://{ip}:{port}/{version}/".format(
            ip=ip,
            port=CONF.haproxy_amphora.bind_port,
            version=API_VERSION)

    def request(self, method, amp, path='/', **kwargs):
        LOG.debug("request url %s", path)
        _request = getattr(self.session, method.lower())
        _url = self._base_url(amp.lb_network_ip) + path
        LOG.debug("request url " + _url)
        timeout_tuple = (CONF.haproxy_amphora.rest_request_conn_timeout,
                         CONF.haproxy_amphora.rest_request_read_timeout)
        reqargs = {
            'url': _url,
            'timeout': timeout_tuple, }
        reqargs.update(kwargs)
        headers = reqargs.setdefault('headers', {})

        headers['User-Agent'] = OCTAVIA_API_CLIENT
        # Keep retrying
        for a in range(CONF.haproxy_amphora.connection_max_retries):
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore",
                        message="A true SSLContext object is not available"
                    )
                    r = _request(**reqargs)
                LOG.debug("Connected to amphora. Response: {resp}".format(
                    resp=r))
                return r
            except (requests.ConnectionError, requests.Timeout):
                LOG.warning(_LW("Could not connect to instance. Retrying."))
                time.sleep(CONF.haproxy_amphora.connection_retry_interval)

        LOG.error(_LE("Connection retries (currently set to %s) "
                      "exhausted.  The amphora is unavailable."),
                  CONF.haproxy_amphora.connection_max_retries)
        raise driver_except.TimeOutException()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This module must import all modules from the GBP repo that define
# SQLAlchemy database models using Neutron's
# neutron_lib.db.model_base.BASEV2 base class. It also must itself be
# imported, directly or indirectly, by every test module for which
# Neutron's neutron.tests.unit.testlib_api.SqlFixture, or any derived
# class, manages the creation of the DB schema prior to running tests
# and the clearing of DB tables between tests.

from gbpservice.neutron.db.grouppolicy.extensions import (  # noqa
    apic_allowed_vm_name_db,
    apic_auto_ptg_db,
    apic_intra_ptg_db,
    apic_reuse_bd_db,
    apic_segmentation_label_db,
    group_proxy_db
)
from gbpservice.neutron.db.grouppolicy import (  # noqa
    group_policy_db,
    group_policy_mapping_db
)
from gbpservice.neutron.db import (  # noqa
    implicitsubnetpool_db,
    servicechain_db
)
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (  # noqa
    data_migrations,
    db,
    extension_db
)
from gbpservice.neutron.services.grouppolicy.drivers import (  # noqa
    chain_mapping,
    implicit_policy,
    nsp_manager,
    resource_mapping
)
from gbpservice.neutron.services.grouppolicy.drivers.cisco.apic import (  # noqa
    apic_mapping
)
from gbpservice.neutron.services.servicechain.plugins.ncp import (  # noqa
    model
)
from gbpservice.neutron.services.servicechain.plugins.ncp.node_drivers import (  # noqa
    heat_node_driver,
    nfp_node_driver
)
from gbpservice.neutron.tests.unit.plugins.ml2plus.drivers import (  # noqa
    extension_test
)
from gbpservice.neutron.tests.unit.services.grouppolicy.drivers import (  # noqa
    extension_test
)
from networking_sfc.db import flowclassifier_db  # noqa
from networking_sfc.db import sfc_db  # noqa

# Note that the models in gbpservice.nfp.orchestrator.db.nfp_db_model
# are managed by a separate fixture, so are not imported here.



# Copyright (c) 2014 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import re

from neutron import context as nctx
from neutron.extensions import portbindings
from neutron import manager
from neutron.plugins.ml2 import driver_api as api
from neutron_lib import constants as n_constants
from opflexagent import constants as ofcst
from oslo_log import log
from oslo_utils import importutils

from gbpservice._i18n import _LW
from gbpservice.neutron.services.grouppolicy.drivers.cisco.apic import (
    apic_mapping as amap)
from gbpservice.neutron.services.grouppolicy.drivers.cisco.apic import (
    nova_client as nclient)


LOG = log.getLogger(__name__)

# TODO(tbachman) Find a good home for these
AGENT_TYPE_DVS = 'DVS agent'
VIF_TYPE_DVS = 'dvs'
DVS_AGENT_KLASS = 'networking_vsphere.common.dvs_agent_rpc_api.DVSClientAPI'


class APICMechanismGBPDriver(api.MechanismDriver):

    def __init__(self):
        super(APICMechanismGBPDriver, self).__init__()
        self._dvs_notifier = None
        self._apic_allowed_vm_name_driver = None

    def _agent_bind_port(self, context, agent_list, bind_strategy):
        """Attempt port binding per agent.

           Perform the port binding for a given agent.
           Returns True if bound successfully.
        """
        for agent in agent_list:
            LOG.debug("Checking agent: %s", agent)
            if agent['alive']:
                for segment in context.segments_to_bind:
                    if bind_strategy(context, segment, agent):
                        LOG.debug("Bound using segment: %s", segment)
                        return True
            else:
                LOG.warning(_LW("Refusing to bind port %(pid)s to dead agent: "
                                "%(agent)s"),
                            {'pid': context.current['id'], 'agent': agent})
        return False

    def bind_port(self, context):
        """Get port binding per host.

           This is similar to the one defined in the
           AgentMechanismDriverBase class, but is modified
           to support multiple L2 agent types (DVS and OpFlex).
        """
        port = context.current
        LOG.debug("Attempting to bind port %(port)s on "
                  "network %(network)s",
                  {'port': port['id'],
                   'network': context.network.current['id']})
        vnic_type = port.get(portbindings.VNIC_TYPE,
                             portbindings.VNIC_NORMAL)
        if vnic_type not in [portbindings.VNIC_NORMAL]:
            LOG.debug("Refusing to bind due to unsupported vnic_type: %s",
                      vnic_type)
            return

        if port['device_owner'].startswith('compute:'):
            # enforce the allowed_vm_names rules if possible
            if (port['device_id'] and self.apic_allowed_vm_name_driver):
                ptg, pt = self.apic_gbp._port_id_to_ptg(
                    context._plugin_context, port['id'])
                if ptg is None:
                    LOG.warning(_LW("PTG for port %s does not exist"),
                                port['id'])
                    return
                l2p = self.apic_gbp._get_l2_policy(context._plugin_context,
                                                   ptg['l2_policy_id'])
                l3p = self.apic_gbp.gbp_plugin.get_l3_policy(
                    context._plugin_context, l2p['l3_policy_id'])

                ok_to_bind = True
                if l3p.get('allowed_vm_names'):
                    ok_to_bind = False
                    vm = nclient.NovaClient().get_server(port['device_id'])
                    for allowed_vm_name in l3p['allowed_vm_names']:
                        match = re.search(allowed_vm_name, vm.name)
                        if match:
                            ok_to_bind = True
                            break
                if not ok_to_bind:
                    LOG.warning(_LW("Failed to bind the port due to "
                                    "allowed_vm_names rules %(rules)s "
                                    "for VM: %(vm)s"),
                                {'rules': l3p['allowed_vm_names'],
                                 'vm': vm.name})
                    return

            # Attempt to bind ports for DVS agents for nova-compute daemons
            # first. This allows having network agents (dhcp, metadata)
            # that typically run on a network node using an OpFlex agent to
            # co-exist with nova-compute daemons for ESX, which host DVS
            # agents.
            agent_list = context.host_agents(AGENT_TYPE_DVS)
            if self._agent_bind_port(context, agent_list, self._bind_dvs_port):
                return

        # It either wasn't a DVS binding, or there wasn't a DVS
        # agent on the binding host (could be the case in a hybrid
        # environment supporting KVM and ESX compute). Go try for
        # OpFlex agents.
        agent_list = context.host_agents(ofcst.AGENT_TYPE_OPFLEX_OVS)
        self._agent_bind_port(context, agent_list, self._bind_opflex_port)

    def _bind_dvs_port(self, context, segment, agent):
        """Populate VIF type and details for DVS VIFs.

           For DVS VIFs, provide the portgroup along
           with the security groups setting
        """
        if self._check_segment_for_agent(segment, agent):
            port = context.current
            # We only handle details for ports that are PTs in PTGs
            ptg, pt = self.apic_gbp._port_id_to_ptg(context._plugin_context,
                                                    port['id'])
            if ptg is None:
                LOG.warning(_LW("PTG for port %s does not exist"), port['id'])
                return False
            mapper = self.apic_gbp.name_mapper
            ptg_name = mapper.policy_target_group(context, ptg)
            network_id = port.get('network_id')
            network = self.apic_gbp._get_network(context._plugin_context,
                                                 network_id)
            project_name = self.apic_gbp._tenant_by_sharing_policy(network)
            apic_tenant_name = self.apic_gbp.apic_manager.apic.fvTenant.name(
                project_name)
            profile = self.apic_gbp.apic_manager.app_profile_name
            # Use default security groups from MD
            vif_details = {portbindings.CAP_PORT_FILTER: False}
            vif_details['dvs_port_group_name'] = (apic_tenant_name +
                                             '|' + str(profile) +
                                             '|' + str(ptg_name))
            currentcopy = copy.copy(context.current)
            currentcopy['portgroup_name'] = (
                vif_details['dvs_port_group_name'])
            booked_port_info = None
            if self.dvs_notifier:
                booked_port_info = self.dvs_notifier.bind_port_call(
                    currentcopy,
                    context.network.network_segments,
                    context.network.current,
                    context.host
                )
            if booked_port_info:
                vif_details['dvs_port_key'] = booked_port_info['key']
            context.set_binding(segment[api.ID],
                                VIF_TYPE_DVS, vif_details,
                                n_constants.PORT_STATUS_ACTIVE)
            return True
        else:
            return False

    def _bind_opflex_port(self, context, segment, agent):
        """Populate VIF type and details for OpFlex VIFs.

           For OpFlex VIFs, we just report the OVS VIF type,
           along with security groups setting, which were
           set when this mechanism driver was instantiated.
        """
        if self._check_segment_for_agent(segment, agent):
            context.set_binding(segment[api.ID],
                                portbindings.VIF_TYPE_OVS,
                                {portbindings.CAP_PORT_FILTER: False,
                                 portbindings.OVS_HYBRID_PLUG: False})
            return True
        else:
            return False

    def _check_segment_for_agent(self, segment, agent):
        """Check support for OpFlex type segments.

           The agent has the ability to limit the segments in OpFlex
           networks by specifying the mappings in their config. If no
           mapping is specifified, then all OpFlex segments are
           supported.
        """
        network_type = segment[api.NETWORK_TYPE]
        if network_type == ofcst.TYPE_OPFLEX:
            opflex_mappings = agent['configurations'].get('opflex_networks')
            LOG.debug("Checking segment: %(segment)s "
                      "for physical network: %(mappings)s ",
                      {'segment': segment, 'mappings': opflex_mappings})
            return (opflex_mappings is None or
                    segment[api.PHYSICAL_NETWORK] in opflex_mappings)
        elif network_type == 'local':
            return True
        else:
            return False

    def initialize(self):
        super(APICMechanismGBPDriver, self).initialize()
        self._apic_gbp = None

    @property
    def apic_gbp(self):
        if not self._apic_gbp:
            self._apic_gbp = manager.NeutronManager.get_service_plugins()[
                'GROUP_POLICY'].policy_driver_manager.policy_drivers[
                'apic'].obj
        return self._apic_gbp

    @property
    def dvs_notifier(self):
        if not self._dvs_notifier:
            try:
                self._dvs_notifier = importutils.import_object(
                    DVS_AGENT_KLASS,
                    nctx.get_admin_context_without_session()
                )
            except ImportError:
                self._dvs_notifier = None
        return self._dvs_notifier

    @property
    def apic_allowed_vm_name_driver(self):
        if self._apic_allowed_vm_name_driver is False:
            return False
        if not self._apic_allowed_vm_name_driver:
            ext_drivers = (self.apic_gbp.gbp_plugin.extension_manager.
                           ordered_ext_drivers)
            for driver in ext_drivers:
                if 'apic_allowed_vm_name' == driver.name:
                    self._apic_allowed_vm_name_driver = driver.obj
                    break
        if not self._apic_allowed_vm_name_driver:
            self._apic_allowed_vm_name_driver = False
        return self._apic_allowed_vm_name_driver

    def create_port_postcommit(self, context):
        self.apic_gbp.process_port_added(context)

    def update_port_postcommit(self, context):
        self.apic_gbp.process_port_changed(context)
        port = context.current
        if (port.get('binding:vif_details') and
                port['binding:vif_details'].get('dvs_port_group_name')) and (
                self.dvs_notifier):
            self.dvs_notifier.update_postcommit_port_call(
                context.current,
                context.original,
                context.network.network_segments[0],
                context.host
            )

    def delete_port_precommit(self, context):
        self.apic_gbp.process_pre_port_deleted(context)

    def delete_port_postcommit(self, context):
        self.apic_gbp.process_port_deleted(context)
        port = context.current
        if (port.get('binding:vif_details') and
                port['binding:vif_details'].get('dvs_port_group_name')) and (
                self.dvs_notifier):
            self.dvs_notifier.delete_port_call(
                context.current,
                context.original,
                context.network.network_segments[0],
                context.host
            )

    def update_subnet_postcommit(self, context):
        self.apic_gbp.process_subnet_changed(context._plugin_context,
                                             context.original, context.current)

    def create_subnet_postcommit(self, context):
        if not context.current['name'].startswith(amap.APIC_OWNED):
            self.apic_gbp.process_subnet_added(context._plugin_context,
                                               context.current)

    def delete_subnet_postcommit(self, context):
        self.apic_gbp.process_subnet_deleted(context._plugin_context,
                                             context.current)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from abc import abstractmethod
import six


def poll_event_desc(*args, **kwargs):
    """Poll Event Decorator.

    NFP modules can define the poll handlers using
    this decorator.
    """
    def decorator(f):
        f._desc = True
        f._spacing = kwargs.pop('spacing', 0)
        f._event = kwargs.pop('event', None)
        return f

    return decorator

"""Meta class. """


class _Meta(type):

    def __init__(cls, names, bases, dict_):
        """Metaclass that allows us to collect decorated periodic tasks."""
        super(_Meta, cls).__init__(names, bases, dict_)

        try:
            cls._poll_desc_table = dict(cls._poll_desc_table)
        except AttributeError:
            cls._poll_desc_table = {}

        for value in cls.__dict__.values():
            if getattr(value, '_desc', False):
                desc = value
                cls._poll_desc_table[desc._event] = desc

"""Base class for nfp event handlers.

Nfp modules derive and implement event handlers
of this class.
"""


@six.add_metaclass(_Meta)
class NfpEventHandler(object):
    # __metaclass__ = ABCMeta

    def __init__(self):
        super(NfpEventHandler, self).__init__()

    def get_poll_desc_table(self):
        return self._poll_desc_table

    @abstractmethod
    def handle_event(self, event):
        """To handle an event.

        :param event: Object of 'Event' class.

        Returns: None
        """
        pass

    @abstractmethod
    def handle_poll_event(self, event):
        """To handle a poll event.

        Core framework will inovke this method of event handler
        when an event timesout.

        :param event: Object of 'Event' class.

        Returns: {'poll':True/False, 'event':<Updated event>}
            'poll': To repoll for the event.
            'event': Updated event, if not passed core will
                repoll on the old event.
        """
        pass

    @abstractmethod
    def event_cancelled(self, event, reason):
        """Notifies that an event is cancelled by core.

        Event could get cancelled,
            a) Event expired. Module can set lifetime for
                an event. If event is not complete with in
                the time, it is auto expired by core.
            b) Event max timedout. Module can set max number
                of times to poll for an event. Event is cancelled
                after the max times.

        :param event: Cancelled event. Object of 'Event' class.
        :param reason: Reason for cancellation. String.

        Returns: None
        """
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy

from gbpservice._i18n import _LI
from gbpservice.contrib.nfp.config_orchestrator.common import common
from gbpservice.contrib.nfp.config_orchestrator.common import lbv2_constants
from gbpservice.nfp.common import constants as const
from gbpservice.nfp.common import data_formatter as df
from gbpservice.nfp.common import utils
from gbpservice.nfp.core import context as module_context
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import transport

from neutron_lbaas.common import cert_manager
from neutron_lbaas.common.tls_utils import cert_parser
from neutron_lbaas.db.loadbalancer import loadbalancer_dbv2
from neutron_lbaas.extensions import loadbalancerv2

from oslo_log import helpers as log_helpers
import oslo_messaging as messaging

LOG = nfp_logging.getLogger(__name__)

"""
RPC handler for Loadbalancer service
"""


class Lbv2Agent(loadbalancer_dbv2.LoadBalancerPluginDbv2):
    target = messaging.Target(version=const.LOADBALANCERV2_RPC_API_VERSION)

    def __init__(self, conf, sc):
        super(Lbv2Agent, self).__init__()
        self._conf = conf
        self._sc = sc
        self._cert_manager_plugin = cert_manager.get_backend()
        self._db_inst = super(Lbv2Agent, self)

    def _filter_service_info_with_resource(self, lb_db, core_db):
        updated_db = {'subnets': [],
                      'ports': []}
        for lb in lb_db['loadbalancers']:
            lb_port_id = lb['vip_port_id']
            lb_subnet_id = lb['vip_subnet_id']
            for subnet in core_db['subnets']:
                if subnet['id'] == lb_subnet_id:
                    updated_db['subnets'].append(subnet)
            for port in core_db['ports']:
                if port['id'] == lb_port_id:
                    updated_db['ports'].append(port)
        lb_db.update(updated_db)
        return lb_db

    def _to_api_dict(self, objs):
        ret_list = []
        for obj in objs:
            ret_list.append(obj.to_api_dict())
        return ret_list

    def _get_core_context(self, context, tenant_id):
        filters = {'tenant_id': [tenant_id]}
        core_context_dict = common.get_core_context(context,
                                                    filters,
                                                    self._conf)
        del core_context_dict['routers']
        return core_context_dict

    def _get_lb_context(self, context, filters):
        args = {'context': context, 'filters': filters}
        db_data = super(Lbv2Agent, self)
        return {'loadbalancers': self._to_api_dict(
            db_data.get_loadbalancers(**args)),
            'listeners': self._to_api_dict(
            db_data.get_listeners(**args)),
            'pools': self._to_api_dict(
            db_data.get_pools(**args)),
            'pool_members': self._to_api_dict(
            db_data.get_pool_members(**args)),
            'healthmonitors': self._to_api_dict(
            db_data.get_healthmonitors(**args))}

    def _context(self, **kwargs):
        context = kwargs.get('context')
        if context.is_admin:
            kwargs['tenant_id'] = context.tenant_id
        core_db = self._get_core_context(context, kwargs['tenant_id'])
        # REVISIT(jiahao): _get_lb_context() fails for flavor_id, disable it
        # for now. Sent the whole core_db to configurator
        # lb_db = self._get_lb_context(**kwargs)
        # db = self._filter_service_info_with_resource(lb_db, core_db)
        db = core_db
        return db

    def _prepare_resource_context_dicts(self, **kwargs):
        # Prepare context_dict
        context = kwargs.get('context')
        context_resource_data = kwargs.pop('context_resource_data')
        ctx_dict = context.to_dict()
        # Collecting db entry required by configurator.
        # Addind service_info to neutron context and sending
        # dictionary format to the configurator.
        db = self._context(**kwargs)
        rsrc_ctx_dict = copy.deepcopy(ctx_dict)
        rsrc_ctx_dict.update({'service_info': db})
        rsrc_ctx_dict.update({'resource_data': context_resource_data})
        return ctx_dict, rsrc_ctx_dict

    def _data_wrapper(self, context, tenant_id, name, reason, nf, **kwargs):
        nfp_context = {}
        description = ast.literal_eval((nf['description'].split('\n'))[1])
        description.update({'tenant_id': tenant_id})
        context_resource_data = df.get_network_function_info(
            description, const.LOADBALANCERV2)
        # REVISIT(dpak): We need to avoid resource description
        # dependency in OTC and instead use neutron context description.
        if name.lower() == 'loadbalancer':
            lb_id = kwargs['loadbalancer']['id']
            kwargs['loadbalancer'].update({'description': str(description)})
            nfp_context = {'network_function_id': nf['id'],
                           'loadbalancer_id': kwargs['loadbalancer']['id']}
        elif name.lower() == 'listener':
            lb_id = kwargs['listener'].get('loadbalancer_id')
            kwargs['listener']['description'] = str(description)
        elif name.lower() == 'pool':
            lb_id = kwargs['pool'].get('loadbalancer_id')
            kwargs['pool']['description'] = str(description)
        elif name.lower() == 'member':
            pool = kwargs['member'].get('pool')
            if pool:
                lb_id = pool.get('loadbalancer_id')
            kwargs['member']['description'] = str(description)
        elif name.lower() == 'healthmonitor':
            pool = kwargs['healthmonitor'].get('pool')
            if pool:
                lb_id = pool.get('loadbalancer_id')
            kwargs['healthmonitor']['description'] = str(description)
        else:
            kwargs[name.lower()].update({'description': str(description)})
            lb_id = kwargs[name.lower()].get('loadbalancer_id')

        args = {'tenant_id': tenant_id,
                'lb_id': lb_id,
                'context': context,
                'description': str(description),
                'context_resource_data': context_resource_data}

        ctx_dict, rsrc_ctx_dict = self._prepare_resource_context_dicts(**args)
        service_vm_context = utils.get_service_vm_context(
                                                description['service_vendor'])
        nfp_context.update({'neutron_context': ctx_dict,
                            'requester': 'nas_service',
                            'logging_context':
                                module_context.get()['log_context'],
                            'service_vm_context': service_vm_context})
        resource_type = 'loadbalancerv2'
        resource = name
        resource_data = {'neutron_context': rsrc_ctx_dict}
        resource_data.update(**kwargs)
        body = common.prepare_request_data(nfp_context, resource,
                                           resource_type, resource_data,
                                           description['service_vendor'])
        return body

    def _post(self, context, tenant_id, name, nf, **kwargs):
        body = self._data_wrapper(context, tenant_id, name,
                                  'CREATE', nf, **kwargs)
        transport.send_request_to_configurator(self._conf,
                                               context, body, "CREATE")

    def _put(self, context, tenant_id, name, nf, **kwargs):
        body = self._data_wrapper(context, tenant_id, name,
                                  'UPDATE', nf, **kwargs)
        transport.send_request_to_configurator(self._conf,
                                               context, body, "UPDATE")

    def _delete(self, context, tenant_id, name, nf, **kwargs):
        body = self._data_wrapper(context, tenant_id, name,
                                  'DELETE', nf, **kwargs)
        transport.send_request_to_configurator(self._conf,
                                               context, body, "DELETE")

    def _fetch_nf_from_resource_desc(self, desc):
        desc_dict = ast.literal_eval(desc)
        nf_id = desc_dict['network_function_id']
        return nf_id

    def _get_primary_cn(self, tls_cert):
        """Returns primary CN for Certificate."""
        return cert_parser.get_host_names(tls_cert.get_certificate())['cn']

    @staticmethod
    def _get_listeners_dict_list(resource_type, resource_dict):
        if resource_type.lower() == 'loadbalancer':
            listeners = resource_dict['listeners']
        elif resource_type.lower() == 'listener':
            listeners = [resource_dict]
        elif resource_type.lower() == 'pool':
            listeners = resource_dict['listeners']
        elif resource_type.lower() == 'member':
            listeners = resource_dict['pool']['listeners']
        elif resource_type.lower() == 'healthmonitor':
            listeners = resource_dict['pool']['listeners']
        else:
            listeners = []

        return listeners

    def _update_tls_cert(self, resource_type, resource_dict):
        listeners = self._get_listeners_dict_list(resource_type, resource_dict)
        for listener in listeners:
            if listener['protocol'] != \
                    lbv2_constants.PROTOCOL_TERMINATED_HTTPS:
                continue
            cert_mgr = self._cert_manager_plugin.CertManager()
            lb_id = listener.get('loadbalancer_id')
            tenant_id = listener.get('tenant_id')

            def get_cert(cont_id):
                try:
                    cert_cont = cert_mgr.get_cert(
                        project_id=tenant_id,
                        cert_ref=cont_id,
                        resource_ref=cert_mgr.get_service_url(lb_id),
                        check_only=True
                    )
                    return cert_cont
                except Exception as e:
                    if hasattr(e, 'status_code') and e.status_code == 404:
                        raise loadbalancerv2.TLSContainerNotFound(
                            container_id=cont_id)
                    else:
                        # Could be a keystone configuration error...
                        raise loadbalancerv2.CertManagerError(
                            ref=cont_id, reason=e.message
                        )

            def build_container_dict(cont_id, cert_cont):
                return {
                    "id": cont_id,
                    "primary_cn": self._get_primary_cn(cert_cont),
                    "private_key": cert_cont.get_private_key(),
                    "certificate": cert_cont.get_certificate(),
                    "intermediates": cert_cont.get_intermediates()
                }

            if not listener['default_tls_container_id']:
                raise loadbalancerv2.TLSDefaultContainerNotSpecified()
            else:
                container_id = listener['default_tls_container_id']
                cert_container = get_cert(container_id)
                container_dict = \
                    build_container_dict(container_id, cert_container)
                listener["default_tls_container"] = container_dict

            for container in listener.get("sni_containers"):
                container_id = container["tls_container_id"]
                cert_container = get_cert(container_id)
                container_dict = \
                    build_container_dict(container_id, cert_container)
                container["tls_container"] = container_dict

    # REVISIT(jiahao): Argument allocate_vip and
    # delete_vip_port are not implememnted.
    @log_helpers.log_method_call
    def create_loadbalancer(self, context, loadbalancer, driver_name,
                            allocate_vip=True):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE LOADBALANCER for LB:%(lb)s"),
                 {'lb': loadbalancer})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('loadbalancer', loadbalancer)
        self._post(
            context, loadbalancer['tenant_id'],
            'loadbalancer', nf,
            loadbalancer=loadbalancer, driver_name=driver_name)

    @log_helpers.log_method_call
    def update_loadbalancer(self, context, old_loadbalancer, loadbalancer):
        nfp_context = module_context.init()
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('loadbalancer', loadbalancer)
        self._put(
            context, loadbalancer['tenant_id'],
            'loadbalancer', nf,
            old_loadbalancer=old_loadbalancer, loadbalancer=loadbalancer)

    @log_helpers.log_method_call
    def delete_loadbalancer(self, context, loadbalancer,
                            delete_vip_port=True):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE LOADBALANCER for LB:"
                     "%(lb)s"), {'lb': loadbalancer})
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('loadbalancer', loadbalancer)
        self._delete(
            context, loadbalancer['tenant_id'],
            'loadbalancer', nf, loadbalancer=loadbalancer)

    @log_helpers.log_method_call
    def create_listener(self, context, listener):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE LISTENER for Listener:%(listener)s"),
                 {'listener': listener})
        loadbalancer = listener['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('listener', listener)
        self._post(
            context, listener['tenant_id'],
            'listener', nf, listener=listener)

    @log_helpers.log_method_call
    def update_listener(self, context, old_listener, listener):
        nfp_context = module_context.init()
        loadbalancer = listener['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('listener', listener)
        self._put(
            context, listener['tenant_id'],
            'listener', nf, old_listener=old_listener, listener=listener)

    @log_helpers.log_method_call
    def delete_listener(self, context, listener):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE LISTENER for Listener:%(listener)s"),
                 {'listener': listener})
        loadbalancer = listener['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('listener', listener)
        self._delete(
            context, listener['tenant_id'],
            'listener', nf, listener=listener)

    @log_helpers.log_method_call
    def create_pool(self, context, pool):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE POOL for Pool:%(pool)s"),
                 {'pool': pool})
        loadbalancer = pool['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('pool', pool)
        self._post(
            context, pool['tenant_id'],
            'pool', nf, pool=pool)

    @log_helpers.log_method_call
    def update_pool(self, context, old_pool, pool):
        nfp_context = module_context.init()
        loadbalancer = pool['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('pool', pool)
        self._put(
            context, pool['tenant_id'],
            'pool', nf, old_pool=old_pool, pool=pool)

    @log_helpers.log_method_call
    def delete_pool(self, context, pool):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE POOL for Pool:%(pool)s"),
                 {'pool': pool})
        loadbalancer = pool['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('pool', pool)
        self._delete(
            context, pool['tenant_id'],
            'pool', nf, pool=pool)

    @log_helpers.log_method_call
    def create_member(self, context, member):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE MEMBER for Member:%(member)s"),
                 {'member': member})
        loadbalancer = member['pool']['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('member', member)
        self._post(
            context, member['tenant_id'],
            'member', nf, member=member)

    @log_helpers.log_method_call
    def update_member(self, context, old_member, member):
        nfp_context = module_context.init()
        loadbalancer = member['pool']['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('member', member)
        self._put(
            context, member['tenant_id'],
            'member', nf, old_member=old_member, member=member)

    @log_helpers.log_method_call
    def delete_member(self, context, member):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE MEMBER for Member:%(member)s"),
                 {'member': member})
        loadbalancer = member['pool']['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('member', member)
        self._delete(
            context, member['tenant_id'],
            'member', nf, member=member)

    @log_helpers.log_method_call
    def create_healthmonitor(self, context, healthmonitor):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC CREATE HEALTH MONITOR for HM:%(hm)s"),
                 {'hm': healthmonitor})
        loadbalancer = healthmonitor['pool']['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('healthmonitor', healthmonitor)
        self._post(
            context, healthmonitor['tenant_id'],
            'healthmonitor', nf, healthmonitor=healthmonitor)

    @log_helpers.log_method_call
    def update_healthmonitor(self, context, old_healthmonitor, healthmonitor):
        nfp_context = module_context.init()
        loadbalancer = healthmonitor['pool']['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('healthmonitor', healthmonitor)
        self._put(
            context, healthmonitor['tenant_id'],
            'healthmonitor', nf,
            old_healthmonitor=old_healthmonitor, healthmonitor=healthmonitor)

    @log_helpers.log_method_call
    def delete_healthmonitor(self, context, healthmonitor):
        nfp_context = module_context.init()
        LOG.info(_LI("Received RPC DELETE HEALTH MONITOR for HM:%(hm)s"),
                 {'hm': healthmonitor})
        loadbalancer = healthmonitor['pool']['loadbalancer']
        # Fetch nf_id from description of the resource
        nf_id = self._fetch_nf_from_resource_desc(loadbalancer["description"])
        nfp_context['log_context']['meta_id'] = nf_id
        nf = common.get_network_function_details(context, nf_id)
        self._update_tls_cert('healthmonitor', healthmonitor)
        self._delete(
            context, healthmonitor['tenant_id'],
            'healthmonitor', nf, healthmonitor=healthmonitor)

    # REVISIT(jiahao): L7policy support not implemented
    # disable L7policy
    # def create_l7policy(self, context, l7policy):
    #     self._post(
    #         context, l7policy['tenant_id'],
    #         'l7policy', l7policy=l7policy)
    #
    # def delete_l7policy(self, context, l7policy):
    #     self._delete(
    #         context, l7policy['tenant_id'],
    #         'l7policy', l7policy=l7policy)
    #
    # def create_l7policy_rule(self, context, rule, l7policy_id):
    #     self._post(
    #         context, rule['tenant_id'],
    #         'rule', rule=rule)
    #
    # def delete_l7policy_rule(self, context, rule):
    #     self._delete(
    #         context, rule['tenant_id'],
    #         'rule', rule=rule)
    #
    # def _get_lb_context(self, context, filters):
    #     args = {'context': context, 'filters': filters}
    #     db_data = super(Lbv2Agent, self)
    #     return {'loadbalancers': db_data.get_loadbalancers(**args),
    #             'listeners': db_data.get_listeners(**args),
    #             'pools': db_data.get_pools(**args),
    #             'pool_members': db_data.get_pool_members(**args),
    #             'healthmonitors': db_data.get_healthmonitors(**args),
    #             'l7policies': db_data.get_l7policies(**args),
    #             'l7policy_rules': db_data.get_l7policy_rules(**args)}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import netaddr

from aim import aim_manager
from aim.api import resource as aim_resource
from aim.api import service_graph as aim_sg
from aim import context as aim_context
from aim import utils as aim_utils
from networking_sfc.db import flowclassifier_db as flowc_db
from networking_sfc.db import sfc_db
from networking_sfc.extensions import flowclassifier as flowc_ext
from networking_sfc.extensions import sfc as sfc_ext
from networking_sfc.services.sfc.common import context as sfc_ctx
from networking_sfc.services.sfc.drivers import base
from neutron.callbacks import events
from neutron.callbacks import registry
from neutron.db import models_v2
from neutron import manager
from neutron_lib import constants as n_constants
from oslo_log import log as logging
from sqlalchemy import or_

from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import apic_mapper
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import constants
from gbpservice.neutron.services.grouppolicy.common import exceptions as exc
from gbpservice.neutron.services.sfc.aim import constants as sfc_cts
from gbpservice.neutron.services.sfc.aim import exceptions

PBR_INGR_PREFIX = 'ingr_'
PBR_EGR_PREFIX = 'egr_'
INGRESS = 'ingress'
EGRESS = 'egress'
FLOWC_SRC = 'src'
FLOWC_DST = 'dst'
LOG = logging.getLogger(__name__)
PHYSDOM_TYPE = 'PhysDom'
SUPPORTED_DOM_TYPES = [PHYSDOM_TYPE]
DEFAULT_SUBNETS = ['128.0.0.0/1', '0.0.0.0/1', '8000::/1', '::/1']
MAX_PPGS_PER_CHAIN = 3


class SfcAIMDriverBase(base.SfcDriverBase):
    def delete_port_pair_group(self, context):
        pass

    def create_port_chain(self, context):
        pass

    def create_port_pair(self, context):
        pass

    def create_port_pair_group(self, context):
        pass

    def delete_port_pair(self, context):
        pass

    def delete_port_chain(self, context):
        pass

    def update_port_pair_group(self, context):
        pass

    def update_port_chain(self, context):
        pass

    def update_port_pair(self, context):
        pass


class SfcAIMDriver(SfcAIMDriverBase):
    """SFC Driver mapping for AIM."""

    def initialize(self):
        # TODO(ivar): SFC resource mapping to APIC DNs
        self._core_plugin = None
        self._flowc_plugin = None
        self._l3_plugin = None
        self._sfc_plugin = None
        self._aim_mech_driver = None
        self._aim_flowc_driver = None
        self.name_mapper = apic_mapper.APICNameMapper()
        self.aim = aim_manager.AimManager()
        # We don't care about deletion, that is managed by the database layer
        # (can't delete a flowclassifier if in use).
        for event in [events.PRECOMMIT_UPDATE, events.PRECOMMIT_CREATE]:
            registry.subscribe(self._handle_flow_classifier,
                               constants.GBP_FLOW_CLASSIFIER, event)
        registry.subscribe(self._handle_port_bound, constants.GBP_PORT,
                           events.PRECOMMIT_UPDATE)
        registry.subscribe(self._handle_net_gbp_change,
                           constants.GBP_NETWORK_EPG, events.PRECOMMIT_UPDATE)
        registry.subscribe(self._handle_net_gbp_change,
                           constants.GBP_NETWORK_VRF, events.PRECOMMIT_UPDATE)
        registry.subscribe(self._handle_net_link_change,
                           constants.GBP_NETWORK_LINK, events.PRECOMMIT_UPDATE)

    @property
    def plugin(self):
        if not self._core_plugin:
            self._core_plugin = manager.NeutronManager.get_plugin()
            if not self._core_plugin:
                LOG.error(_("No Core plugin found."))
                raise exc.GroupPolicyDeploymentError()
        return self._core_plugin

    @property
    def flowc_plugin(self):
        if not self._flowc_plugin:
            plugins = manager.NeutronManager.get_service_plugins()
            self._flowc_plugin = plugins.get(flowc_ext.FLOW_CLASSIFIER_EXT)
            if not self._flowc_plugin:
                LOG.error(_("No FlowClassifier service plugin found."))
                raise exc.GroupPolicyDeploymentError()
        return self._flowc_plugin

    @property
    def l3_plugin(self):
        if not self._l3_plugin:
            plugins = manager.NeutronManager.get_service_plugins()
            self._l3_plugin = plugins.get(n_constants.L3)
            if not self._l3_plugin:
                LOG.error(_("No L3 service plugin found."))
                raise exc.GroupPolicyDeploymentError()
        return self._l3_plugin

    @property
    def sfc_plugin(self):
        if not self._sfc_plugin:
            plugins = manager.NeutronManager.get_service_plugins()
            self._sfc_plugin = plugins.get(sfc_ext.SFC_EXT)
            if not self._sfc_plugin:
                LOG.error(_("No SFC service plugin found."))
                raise exc.GroupPolicyDeploymentError()
        return self._sfc_plugin

    @property
    def aim_mech(self):
        if not self._aim_mech_driver:
            try:
                self._aim_mech_driver = (
                    self.plugin.mechanism_manager.mech_drivers['apic_aim'].obj)
            except (KeyError, AttributeError):
                LOG.error(_("No AIM driver found"))
                raise exc.GroupPolicyDeploymentError()
        return self._aim_mech_driver

    @property
    def aim_flowc(self):
        if not self._aim_flowc_driver:
            try:
                self._aim_flowc_driver = (
                    self.flowc_plugin.driver_manager.drivers['aim'].obj)
            except (KeyError, AttributeError):
                LOG.error(_("No AIM flowc driver found"))
                raise exc.GroupPolicyDeploymentError()
        return self._aim_flowc_driver

    def create_port_pair_precommit(self, context):
        """Map Port Pair to AIM model

        A Port Pair by itself doesn't need to generate AIM model at least
        until added to a Port Pair Group.
        :param context:
        :return:
        """
        self._validate_port_pair(context)

    def update_port_pair_precommit(self, context, remap=False):
        self._validate_port_pair(context)
        p_ctx = context._plugin_context
        # Remap the affected groups if needed.
        if remap or self._should_regenerate_pp(context):
            for group in self._get_groups_by_pair_id(p_ctx,
                                                     context.current['id']):
                # Curr and original are identical, so the same object gets
                # remapped.
                g_ctx = sfc_ctx.PortPairGroupContext(context._plugin, p_ctx,
                                                     group, group)
                self.update_port_pair_group_precommit(g_ctx, remap=True)

    def delete_port_pair_precommit(self, context):
        # NOTE(ivar): DB layer prevents port pair deletion when in use by a
        # port pair group.
        pass

    def create_port_pair_group_precommit(self, context):
        """Map port pair group to AIM model

        A Port Pair Group is the equivalent of a Logical Device in AIM.
        :param context:
        :return:
        """
        self._validate_port_pair_group(context)

    def update_port_pair_group_precommit(self, context, remap=False):
        self._validate_port_pair_group(context)
        # Remap Port Chain if needed
        if remap or self._should_regenerate_ppg(context):
            for chain in self._get_chains_by_ppg_ids(context._plugin_context,
                                                     [context.current['id']]):
                c_ctx = sfc_ctx.PortChainContext(
                    context._plugin, context._plugin_context, chain, chain)
                self.update_port_chain_precommit(c_ctx, remap=True)

    def delete_port_pair_group_precommit(self, context):
        # NOTE(ivar): DB layer prevents deletion when used by port chains
        pass

    def create_port_chain_precommit(self, context):
        pc = context.current
        p_ctx = context._plugin_context
        flowcs, ppgs = self._get_pc_flowcs_and_ppgs(p_ctx, pc)
        self._validate_port_chain(context._plugin_context, context.current,
                                  flowcs, ppgs)
        self._map_port_chain(p_ctx, pc, flowcs, ppgs)

    def update_port_chain_precommit(self, context, remap=False):
        p_ctx = context._plugin_context
        flowcs, ppgs = self._get_pc_flowcs_and_ppgs(p_ctx, context.current)
        self._validate_port_chain(context._plugin_context, context.current,
                                  flowcs, ppgs)
        # Regenerate Port Chain Model
        if remap or self._should_regenerate_pc(context):
            o_flowcs, o_ppgs = self._get_pc_flowcs_and_ppgs(p_ctx,
                                                            context.original)
            self._delete_port_chain_mapping(p_ctx, context.original, o_flowcs,
                                            o_ppgs)
            self._map_port_chain(p_ctx, context.current, flowcs, ppgs)

    def delete_port_chain_precommit(self, context):
        p_ctx = context._plugin_context
        flowcs, ppgs = self._get_pc_flowcs_and_ppgs(p_ctx, context.current)
        self._delete_port_chain_mapping(p_ctx, context.current, flowcs, ppgs)

    def _validate_port_pair(self, context):
        # Ports need to belong to distinct networks
        p_ctx = context._plugin_context
        ingress_port = self.plugin.get_port(p_ctx, context.current['ingress'])
        egress_port = self.plugin.get_port(p_ctx, context.current['egress'])
        ingress_net = ingress_port['network_id']
        egress_net = egress_port['network_id']
        if ingress_net == egress_net:
            raise exceptions.PortPairsSameNetwork(id=context.current['id'])
        igress_dom = self.aim_mech._get_port_unique_domain(p_ctx, ingress_port)
        egress_dom = self.aim_mech._get_port_unique_domain(p_ctx, egress_port)
        if igress_dom != egress_dom:
            raise exceptions.PortPairsDifferentDomain(id=context.current['id'])
        if igress_dom == (None, None):
            raise exceptions.PortPairsNoUniqueDomain(id=context.current['id'])
        # Ensure  domain types supported
        if igress_dom[0] not in SUPPORTED_DOM_TYPES:
            raise exceptions.PortPairsUnsupportedDomain(
                id=context.current['id'], doms=SUPPORTED_DOM_TYPES)
        ingress_net = self.plugin.get_network(p_ctx, ingress_net)
        egress_net = self.plugin.get_network(p_ctx, egress_net)
        if ingress_net.get('apic:svi') or egress_net.get(
                'apic:svi') or ingress_net.get(
                'router:external') or egress_net.get('router:external'):
            raise exceptions.ServiceNetworkBadType(id=context.current['id'])

    def _validate_port_pair_group(self, context):
        # Verify all ports are in the same network for each side of the
        # connection
        p_ctx = context._plugin_context
        port_pairs = context._plugin.get_port_pairs(
            p_ctx, filters={'id': context.current['port_pairs']})
        domains = set()
        net_pairs = set()
        for port_pair in port_pairs:
            ingress_port = self.plugin.get_port(p_ctx, port_pair['ingress'])
            egress_port = self.plugin.get_port(p_ctx, port_pair['egress'])
            domains.add(self.aim_mech._get_port_unique_domain(p_ctx,
                                                              ingress_port))
            if len(domains) > 1:
                raise exceptions.PortPairsInPortPairGroupDifferentDomain(
                    id=context.current['id'])
            net_pairs.add((ingress_port['network_id'],
                           egress_port['network_id']))
            if len(net_pairs) > 1:
                raise exceptions.PortPairsDifferentNetworkInGroup(
                    id=context.current['id'])

    def _validate_port_chain(self, p_ctx, pc, flowcs, ppgs):
        # - All networks in play (prov, cons, services) are in the same VRF,
        #   also listen to events to prevent VRF to change after the fact
        #   alternatively, an ERROR status can be raised
        # - TEMPORARY: provider and consumer EPGs are in the same tenant, this
        #   can be removed once contract export is implemented.
        # TODO(ivar): two different chains cannot share left/right networks
        # TODO(ivar): right/left BDs same tenant as provider
        if len(ppgs) > MAX_PPGS_PER_CHAIN:
            raise exceptions.TooManyPPGsPerChainError(maxn=MAX_PPGS_PER_CHAIN)
        vrfs = set()
        for flowc in flowcs:
            provg = self._get_flowc_provider_group(p_ctx, flowc)
            consg = self._get_flowc_consumer_group(p_ctx, flowc)
            provrf = self._get_flowc_provider_vrf(p_ctx, flowc)
            convrf = self._get_flowc_consumer_vrf(p_ctx, flowc)
            vrfs.add(provrf.dn)
            vrfs.add(convrf.dn)
            if provg.tenant_name != consg.tenant_name:
                raise exceptions.FlowClassifierSrcDstNetworksDifferentTenants(
                    id=flowc['id'])
            if len(vrfs) > 1:
                raise exceptions.MultipleVRFsDetectedInPortChain(id=pc['id'])
        nets = set()
        net_count = 0
        for ppg in ppgs:
            lvrf, rvrf = self._get_ppg_left_right_vrfs(p_ctx, ppg)
            # REVISIT(ivar): current ACI limitation, we can't put subsequent
            # nodes of the same chain in the same networks
            lnet, rnet = self._get_ppg_left_right_network_ids(p_ctx, ppg)
            nets.add(lnet)
            nets.add(rnet)
            net_count += 2
            vrfs.add(lvrf.dn)
            vrfs.add(rvrf.dn)
            if len(vrfs) > 1:
                raise exceptions.MultipleVRFsDetectedInPortChain(id=pc['id'])
            if len(nets) < net_count:
                raise exceptions.ConflictingNetworksDetectedInPortChain(
                    id=pc['id'])

    def _map_port_pair_group(self, plugin_context, ppg, tenant):
        session = plugin_context.session
        aim_ctx = aim_context.AimContext(session)
        # Create Logical device model, container for all the PPG port pairs.
        dc = self._get_ppg_device_cluster(session, ppg, tenant)
        mp = self._get_ppg_monitoring_policy(session, ppg, tenant)
        type, domain = self._get_ppg_domain(plugin_context, ppg)
        if not type and not domain:
            raise exceptions.PortPairsNoUniqueDomain(id=ppg['port_pairs'])
        if type == PHYSDOM_TYPE:
            dc.device_type = 'PHYSICAL'
            dc.physical_domain_name = domain
        else:
            dc.device_type = 'VIRTUAL'
            dc.vmm_domain = [{'type': type, 'name': domain}]
        internal_dci = aim_sg.DeviceClusterInterface(
            tenant_name=dc.tenant_name, device_cluster_name=dc.name,
            name=INGRESS, display_name=INGRESS)
        external_dci = aim_sg.DeviceClusterInterface(
            tenant_name=dc.tenant_name, device_cluster_name=dc.name,
            name=EGRESS, display_name=EGRESS)
        # Create 2 PBR rules per PPG, one per direction.
        ipbr = self._get_ppg_service_redirect_policy(session, ppg, INGRESS,
                                                     tenant)
        epbr = self._get_ppg_service_redirect_policy(session, ppg, EGRESS,
                                                     tenant)
        self.aim.create(aim_ctx, dc)
        if mp:
            for pbr in [ipbr, epbr]:
                pbr.monitoring_policy_tenant_name = mp.tenant_name
                pbr.monitoring_policy_name = mp.name
            self.aim.create(aim_ctx, mp)
        # For each port pair, create the corresponding Concrete Devices
        # (represented by the static path of each interface)
        ingress_cdis = []
        egress_cdis = []
        port_pairs = self.sfc_plugin.get_port_pairs(
            plugin_context, filters={'id': ppg['port_pairs']})
        for pp in port_pairs:
            ingress_port = self.plugin.get_port(plugin_context, pp['ingress'])
            egress_port = self.plugin.get_port(plugin_context, pp['egress'])
            pp_id = self.name_mapper.port_pair(session, pp['id'])
            pp_name = aim_utils.sanitize_display_name(pp['name'])
            cd = aim_sg.ConcreteDevice(
                tenant_name=dc.tenant_name, device_cluster_name=dc.name,
                name=pp_id, display_name=pp_name)
            self.aim.create(aim_ctx, cd)
            srhg = None
            if mp:
                # Create health groups
                srhg = aim_sg.ServiceRedirectHealthGroup(
                    tenant_name=dc.tenant_name, name=pp_id,
                    display_name=pp_name)
                self.aim.create(aim_ctx, srhg)
            for p, store in [(ingress_port, ingress_cdis),
                             (egress_port, egress_cdis)]:
                p_id = self.name_mapper.port(session, p['id'])
                p_name = aim_utils.sanitize_display_name(p['name'])
                path, encap, host = self.aim_mech._get_port_static_path_info(
                    plugin_context, p)
                if path is None:
                    LOG.warning("Path not found for Port Pair %s member %s "
                                "Port might be unbound." % (pp['id'], p['id']))
                    path = ''
                    # TODO(ivar): what if encap is None? is it an Opflex port?
                # Create Concrete Device Interface
                cdi = aim_sg.ConcreteDeviceInterface(
                    tenant_name=cd.tenant_name,
                    device_cluster_name=cd.device_cluster_name,
                    device_name=cd.name, name=p_id, display_name=p_name,
                    path=path, host=host)
                cdi = self.aim.create(aim_ctx, cdi)
                store.append((cdi, encap, p, srhg, pp_id))

        for i in range(len(ingress_cdis)):
            icdi, iencap, iport, isrhg, pp_id = ingress_cdis[i]
            ecdi, eencap, eport, esrhg, _ = egress_cdis[i]
            internal_dci.encap = iencap
            external_dci.encap = eencap
            internal_dci.concrete_interfaces.append(icdi.dn)
            external_dci.concrete_interfaces.append(ecdi.dn)
            if iport['fixed_ips']:
                destination = {'ip': iport['fixed_ips'][0]['ip_address'],
                               'mac': iport['mac_address'], 'name': pp_id}
                if isrhg:
                    destination['redirect_health_group_dn'] = isrhg.dn
                ipbr.destinations.append(destination)
            if eport['fixed_ips']:
                destination = {'ip': eport['fixed_ips'][0]['ip_address'],
                               'mac': eport['mac_address'], 'name': pp_id}
                if esrhg:
                    destination['redirect_health_group_dn'] = esrhg.dn
                epbr.destinations.append(destination)

        self.aim.create(aim_ctx, internal_dci)
        self.aim.create(aim_ctx, external_dci)
        self.aim.create(aim_ctx, ipbr)
        self.aim.create(aim_ctx, epbr)

    def _delete_port_pair_group_mapping(self, plugin_context, ppg, tenant):
        # Just delete cascade the DeviceCluster and PBR policies
        session = plugin_context.session
        aim_ctx = aim_context.AimContext(session)
        dc = self._get_ppg_device_cluster(session, ppg, tenant)
        mp = self._get_ppg_monitoring_policy(session, ppg, tenant)
        self.aim.delete(aim_ctx, dc, cascade=True)
        if mp:
            self.aim.delete(aim_ctx, mp, cascade=True)
            for pp in ppg['port_pairs']:
                pp_id = self.name_mapper.port_pair(session, pp)
                srhg = aim_sg.ServiceRedirectHealthGroup(
                    tenant_name=dc.tenant_name, name=pp_id)
                self.aim.delete(aim_ctx, srhg)
        for prefix in [PBR_INGR_PREFIX, PBR_EGR_PREFIX]:
            pbr_id = self.name_mapper.port_pair_group(session, ppg['id'],
                                                      prefix=prefix)
            self.aim.delete(
                aim_ctx, aim_sg.ServiceRedirectPolicy(
                    tenant_name=dc.tenant_name, name=pbr_id), cascade=True)

    def _map_port_chain(self, plugin_context, pc, flowcs, ppgs):
        # Create one DeviceClusterContext per PPG
        p_ctx = plugin_context
        aim_ctx = aim_context.AimContext(p_ctx.session)
        # For each flow classifier, there are as many DeviceClusterContext as
        # the number of nodes in the chain.
        p_tenants = set()
        for flc in flowcs:
            p_group = self._get_flowc_provider_group(plugin_context, flc)
            p_tenant = p_group.tenant_name
            sg = self._get_pc_service_graph(p_ctx.session, pc, p_tenant)
            contract = self._get_flc_contract(p_group, sg)
            subject = aim_resource.ContractSubject(
                tenant_name=contract.tenant_name, contract_name=contract.name,
                name=sg.name, service_graph_name=sg.name,
                bi_filters=[self.aim_mech._any_filter_name])
            if not self.aim.get(aim_ctx, contract):
                self.aim.create(aim_ctx, contract)
                self.aim.create(aim_ctx, subject)
            self._map_flow_classifier(p_ctx, flc, pc, p_tenant)
            # Map device clusters for each flow tenant
            if p_tenant not in p_tenants:
                for ppg in ppgs:
                    dc = self._get_ppg_device_cluster(p_ctx.session, ppg,
                                                      p_tenant)
                    self._map_port_pair_group(plugin_context, ppg, p_tenant)
                    dcc = aim_sg.DeviceClusterContext(
                        tenant_name=sg.tenant_name, contract_name="any",
                        service_graph_name=sg.name, node_name=dc.name,
                        display_name=dc.display_name,
                        device_cluster_name=dc.name,
                        device_cluster_tenant_name=dc.tenant_name)
                    dcc = self.aim.create(aim_ctx, dcc)
                    # Create device context interfaces.
                    left_bd, right_bd = self._get_ppg_left_right_bds(p_ctx,
                                                                     ppg)
                    for conn_name, direction, bd in [
                            ('provider', EGRESS, right_bd),
                            ('consumer', INGRESS, left_bd)]:
                        dci = aim_sg.DeviceClusterInterface(
                            tenant_name=dc.tenant_name,
                            device_cluster_name=dc.name, name=direction)
                        pbr = self._get_ppg_service_redirect_policy(
                            p_ctx.session, ppg, direction, p_tenant)
                        dcic = aim_sg.DeviceClusterInterfaceContext(
                            tenant_name=dcc.tenant_name,
                            contract_name=dcc.contract_name,
                            service_graph_name=dcc.service_graph_name,
                            node_name=dcc.node_name, connector_name=conn_name,
                            display_name=dcc.display_name,
                            bridge_domain_dn=bd.dn,
                            device_cluster_interface_dn=dci.dn,
                            service_redirect_policy_dn=pbr.dn)
                        self.aim.create(aim_ctx, dcic)
                    sg.linear_chain_nodes.append(
                        {'name': dc.name, 'device_cluster_name': dc.name,
                         'device_cluster_tenant_name': dc.tenant_name})
                    # Unsync left-right EPGs
                    for epg in self._get_ppg_left_right_epgs(p_ctx, ppg):
                        self.aim.update(aim_ctx, epg, sync=False)
                # Create only once per tenant
                self.aim.create(aim_ctx, sg)
                p_tenants.add(p_tenant)

    def _delete_port_chain_mapping(self, plugin_context, pc, flowcs, ppgs):
        p_ctx = plugin_context
        session = p_ctx.session
        aim_ctx = aim_context.AimContext(session)
        deleted_ppgs = set()
        for flc in flowcs:
            p_group = self._get_flowc_provider_group(plugin_context, flc)
            tenant = p_group.tenant_name
            for ppg in ppgs:
                key = (tenant, ppg['id'])
                if key not in deleted_ppgs:
                    self._delete_port_pair_group_mapping(p_ctx, ppg, tenant)
                    deleted_ppgs.add(key)
            self._delete_flow_classifier_mapping(p_ctx, flc, pc, tenant)
            sg = self._get_pc_service_graph(p_ctx.session, pc, tenant)
            contract = self._get_flc_contract(p_group, sg)
            self.aim.delete(aim_ctx, contract, cascade=True)
            self.aim.delete(aim_ctx, sg, cascade=True)
            for ppg_id in pc['port_pair_groups']:
                ppg_aid = self.name_mapper.port_pair_group(session, ppg_id)
                dcc = aim_sg.DeviceClusterContext(
                    tenant_name=tenant, contract_name="any",
                    service_graph_name=sg.name, node_name=ppg_aid)
                self.aim.delete(aim_ctx, dcc, cascade=True)
        processed_networks = set()
        # deleted ppgs contains all the ppgs' ID
        processed_ppgs = deleted_ppgs
        for ppg in ppgs:
            for net_id in self._get_ppg_left_right_network_ids(p_ctx, ppg):
                if net_id in processed_networks:
                    continue
                processed_networks.add(net_id)
                # See if there are more chains on these networks
                for group_id in self._get_group_ids_by_network_ids(p_ctx,
                                                                   [net_id]):
                    if group_id in processed_ppgs:
                        # Nothing to do
                        continue
                    processed_ppgs.add(group_id)
                    for chain in self._get_chains_by_ppg_ids(p_ctx,
                                                             [group_id]):
                        if chain['id'] != pc['id']:
                            # This network is in use by some chain, cannot
                            # re-activate EPG
                            break
                    else:
                        # No chain associated to this group ID
                        continue
                    break
                else:
                    # No chain associated to all the groups of this network
                    epg = self.aim_mech._get_epg_by_network_id(p_ctx.session,
                                                               net_id)
                    self.aim.update(aim_ctx, epg, sync=True)

    def _map_flow_classifier(self, plugin_context, flowc, pc, tenant):
        """Map flowclassifier to AIM model

        If source/destination ports are plugged to external networks, create
        AIM external EPGs in the proper L3Outs and set the corresponding
        source/destination ip prefix.

        :param context:
        :return:
        """
        aim_ctx = aim_context.AimContext(plugin_context.session)
        cons_group = self._map_flowc_consumer_group(plugin_context, flowc)
        prov_group = self._map_flowc_provider_group(plugin_context, flowc)
        sg = self._get_pc_service_graph(plugin_context.session, pc, tenant)
        contract = self._get_flc_contract(prov_group, sg)
        # TODO(ivar): if provider/consumer are in different tenants, export
        # the contract
        if contract.name not in cons_group.consumed_contract_names:
            cons_group.consumed_contract_names.append(contract.name)
        if contract.name not in prov_group.provided_contract_names:
            prov_group.provided_contract_names.append(contract.name)
        self.aim.create(aim_ctx, cons_group, overwrite=True)
        self.aim.create(aim_ctx, prov_group, overwrite=True)

    def _map_flowc_network_group(self, plugin_context, net, cidr, flowc,
                                 prefix):
        flc_aid = self._get_external_group_aim_name(plugin_context, flowc,
                                                    prefix)
        aim_ctx = aim_context.AimContext(plugin_context.session)
        cidr = netaddr.IPNetwork(cidr)
        l3out = self.aim_mech._get_svi_net_l3out(net)
        if l3out:
            # Create ExternalNetwork and ExternalSubnet on the proper
            # L3Out. Return the External network
            ext_net = aim_resource.ExternalNetwork(
                tenant_name=l3out.tenant_name, l3out_name=l3out.name,
                name=flc_aid)
            ext_net_db = self.aim.get(aim_ctx, ext_net)
            if not ext_net_db:
                ext_net_db = self.aim.create(aim_ctx, ext_net)
            subnets = [str(cidr)] if cidr.prefixlen != 0 else DEFAULT_SUBNETS
            for sub in subnets:
                ext_sub = aim_resource.ExternalSubnet(
                    tenant_name=ext_net.tenant_name,
                    l3out_name=ext_net.l3out_name,
                    external_network_name=ext_net.name, cidr=sub)
                ext_sub_db = self.aim.get(aim_ctx, ext_sub)
                if not ext_sub_db:
                    self.aim.create(aim_ctx, ext_sub)
            return ext_net_db
        else:
            return self.aim_mech._get_epg_by_network_id(plugin_context.session,
                                                        net['id'])

    def _delete_flow_classifier_mapping(self, plugin_context, flowc, pc,
                                        tenant):
        source_net = self._get_flowc_src_network(plugin_context, flowc)
        dest_net = self._get_flowc_dst_network(plugin_context, flowc)
        self._delete_flowc_network_group_mapping(
            plugin_context, source_net, flowc, pc, tenant, FLOWC_SRC)
        self._delete_flowc_network_group_mapping(
            plugin_context, dest_net, flowc, pc, tenant, FLOWC_DST)

    def _delete_flowc_network_group_mapping(self, plugin_context, net, flowc,
                                            pc, tenant, prefix=''):
        flc_aid = self._get_external_group_aim_name(plugin_context, flowc,
                                                    prefix)
        aim_ctx = aim_context.AimContext(plugin_context.session)
        l3out = self.aim_mech._get_svi_net_l3out(net)
        ext_net = None
        if l3out:
            ext_net = aim_resource.ExternalNetwork(
                tenant_name=l3out.tenant_name, l3out_name=l3out.name,
                name=flc_aid)
            epg = self.aim.get(aim_ctx, ext_net)
        else:
            epg = self.aim.get(aim_ctx, self.aim_mech._get_epg_by_network_id(
                plugin_context.session, net['id']))
        if epg:
            p_group = self._get_flowc_provider_group(plugin_context, flowc)
            sg = self._get_pc_service_graph(plugin_context.session, pc, tenant)
            contract = self._get_flc_contract(p_group, sg)
            try:
                if prefix == FLOWC_SRC:
                    epg.consumed_contract_names.remove(contract.name)
                else:
                    epg.provided_contract_names.remove(contract.name)
            except ValueError:
                LOG.warning("Contract %s not present in EPG %s" %
                            (contract.name, epg))
            else:
                epg = self.aim.create(aim_ctx, epg, overwrite=True)
            if (ext_net and not epg.consumed_contract_names and not
                    epg.provided_contract_names):
                # Only remove external network if completely empty
                self.aim.delete(aim_ctx, epg, cascade=True)

    def _get_chains_by_classifier_id(self, plugin_context, flowc_id):
        context = plugin_context
        with context.session.begin(subtransactions=True):
            chain_ids = [x.portchain_id for x in context.session.query(
                sfc_db.ChainClassifierAssoc).filter_by(
                flowclassifier_id=flowc_id).all()]
            return self.sfc_plugin.get_port_chains(plugin_context,
                                                   filters={'id': chain_ids})

    def _get_chains_by_ppg_ids(self, plugin_context, ppg_ids):
        context = plugin_context
        with context.session.begin(subtransactions=True):
            chain_ids = [x.portchain_id for x in context.session.query(
                sfc_db.ChainGroupAssoc).filter(
                sfc_db.ChainGroupAssoc.portpairgroup_id.in_(ppg_ids)).all()]
            return self.sfc_plugin.get_port_chains(plugin_context,
                                                   filters={'id': chain_ids})

    def _get_groups_by_pair_id(self, plugin_context, pp_id):
        # NOTE(ivar): today, port pair can be associated only to one PPG
        context = plugin_context
        with context.session.begin(subtransactions=True):
            pp_db = self.sfc_plugin._get_port_pair(plugin_context, pp_id)
            if pp_db and pp_db.portpairgroup_id:
                return self.sfc_plugin.get_port_pair_groups(
                    plugin_context, filters={'id': [pp_db.portpairgroup_id]})
        return []

    def _get_group_ids_by_network_ids(self, plugin_context, network_ids):
        session = plugin_context.session
        return [
            x.portpairgroup_id for x in
            session.query(sfc_db.PortPair).join(
                models_v2.Port,
                or_(models_v2.Port.id == sfc_db.PortPair.ingress,
                    models_v2.Port.id == sfc_db.PortPair.egress)).filter(
                models_v2.Port.network_id.in_(network_ids)).all()]

    def _should_regenerate_pp(self, context):
        attrs = [INGRESS, EGRESS, 'name']
        return any(context.current[a] != context.original[a] for a in attrs)

    def _should_regenerate_ppg(self, context):
        attrs = ['port_pairs', 'name']
        param_curr = context.current['port_pair_group_parameters']
        param_orig = context.original['port_pair_group_parameters']
        return (any(context.current[a] != context.original[a] for a in attrs)
                or any(param_curr.get(x) != param_orig.get(x) for x in
                       sfc_cts.AIM_PPG_PARAMS.keys()))

    def _should_regenerate_pc(self, context):
        attrs = ['flow_classifiers', 'port_pair_groups', 'name']
        return any(context.current[a] != context.original[a] for a in attrs)

    def _should_regenerate_flowc(self, context):
        current = context.current
        original = context.original
        l7_curr = current['l7_parameters']
        l7_orig = original['l7_parameters']
        return (
            any(current[x] != original[x] for x in
                sfc_cts.AIM_FLC_PARAMS + ['name'])
            or any(l7_curr[x] != l7_orig[x] for x in
                   sfc_cts.AIM_FLC_L7_PARAMS.keys()))

    def _get_ppg_device_cluster(self, session, ppg, tenant):
        tenant_aid = tenant
        ppg_aid = self.name_mapper.port_pair_group(session, ppg['id'])
        ppg_aname = aim_utils.sanitize_display_name(ppg['name'])
        return aim_sg.DeviceCluster(tenant_name=tenant_aid, name=ppg_aid,
                                    display_name=ppg_aname, managed=False)

    def _get_ppg_monitoring_policy(self, session, ppg, tenant):
        tenant_aid = tenant
        ppg_aid = self.name_mapper.port_pair_group(session, ppg['id'])
        ppg_aname = aim_utils.sanitize_display_name(ppg['name'])
        result = None
        ppg_params = ppg.get('port_pair_group_parameters', {})
        monitoring_policy = ppg_params.get('healthcheck_type')
        if monitoring_policy:
            kwargs = {'type': monitoring_policy}
            if 'healthcheck_frequency' in ppg_params:
                kwargs['frequency'] = ppg_params['healthcheck_frequency']
            if 'healthcheck_tcp_port' in ppg_params:
                kwargs['tcp_port'] = ppg_params['healthcheck_tcp_port']
            result = aim_sg.ServiceRedirectMonitoringPolicy(
                tenant_name=tenant_aid, name=ppg_aid,
                display_name=ppg_aname, **kwargs)
        return result

    def _get_ppg_domain(self, plugin_context, ppg):
        pp = self.sfc_plugin.get_port_pair(plugin_context,
                                           ppg['port_pairs'][0])
        ingress_port = self.plugin.get_port(plugin_context, pp['ingress'])
        return self.aim_mech._get_port_unique_domain(plugin_context,
                                                     ingress_port)

    def _get_pc_service_graph(self, session, pc, tenant):
        tenant_aid = tenant
        pc_aid = self.name_mapper.port_chain(session, pc['id'])
        pc_aname = aim_utils.sanitize_display_name(pc['name'])
        return aim_sg.ServiceGraph(tenant_name=tenant_aid, name=pc_aid,
                                   display_name=pc_aname)

    def _get_flc_contract(self, provider, graph):
        tenant_id = provider.tenant_name
        # NOTE(ivar): For scalability purposes, we want to limit the number
        # of contracts by creating one per provider/graph pair.
        # In order to keep the contract name unique, hashing the provider and
        # graph names (which are already unique) together seems like the
        # safest choice, considering that truncating those names to fit 64
        # bytes would cause a huge loss of entropy. This is especially true
        # for SVI networks, where the provider group can use as much as
        # 23 bytes only for the prefix ([network cidr] + _net_)
        return aim_resource.Contract(
            tenant_name=tenant_id,
            name=self._generate_contract_name(provider.name, graph.name),
            display_name=graph.display_name)

    def _generate_contract_name(self, prov_name, graph_name):
        return hashlib.sha256(prov_name + graph_name).hexdigest()

    def _get_ppg_service_redirect_policy(self, session, ppg, direction,
                                         tenant):
        if direction == INGRESS:
            prfx = PBR_INGR_PREFIX
        elif direction == EGRESS:
            prfx = PBR_EGR_PREFIX
        dc = self._get_ppg_device_cluster(session, ppg, tenant)
        pbr_id = self.name_mapper.port_pair_group(session, ppg['id'],
                                                  prefix=prfx)
        return aim_sg.ServiceRedirectPolicy(tenant_name=dc.tenant_name,
                                            name=pbr_id)

    def _get_ppg_left_right_network_ids(self, plugin_context, ppg):
        pps = self.sfc_plugin.get_port_pairs(plugin_context,
                                             filters={'id': ppg['port_pairs']})
        for pp in pps:
            ingress = self.plugin.get_port(plugin_context, pp['ingress'])
            egress = self.plugin.get_port(plugin_context, pp['egress'])
            return ingress['network_id'], egress['network_id']

    def _get_ppg_left_right_epgs(self, plugin_context, ppg):
        leftn, rightn = self._get_ppg_left_right_network_ids(plugin_context,
                                                             ppg)
        ingress_epg = self.aim_mech._get_epg_by_network_id(
            plugin_context.session, leftn)
        egress_epg = self.aim_mech._get_epg_by_network_id(
            plugin_context.session, rightn)
        # Every port pair will return the same result
        return ingress_epg, egress_epg

    def _get_ppg_left_right_bds(self, plugin_context, ppg):
        leftn, rightn = self._get_ppg_left_right_network_ids(plugin_context,
                                                             ppg)
        ingress_bd = self.aim_mech._get_bd_by_network_id(
            plugin_context.session, leftn)
        egress_bd = self.aim_mech._get_bd_by_network_id(
            plugin_context.session, rightn)
        # Every port pair will return the same result
        return ingress_bd, egress_bd

    def _get_ppg_left_right_vrfs(self, plugin_context, ppg):
        leftn, rightn = self._get_ppg_left_right_network_ids(plugin_context,
                                                             ppg)
        leftn = self.plugin.get_network(plugin_context, leftn)
        rightn = self.plugin.get_network(plugin_context, rightn)
        ingress_vrf = self.aim_mech._get_vrf_by_network(plugin_context.session,
                                                        leftn)
        egress_vrf = self.aim_mech._get_vrf_by_network(plugin_context.session,
                                                       rightn)
        # Every port pair will return the same result
        return ingress_vrf, egress_vrf

    def _handle_flow_classifier(self, rtype, event, trigger, driver_context,
                                **kwargs):
        if event == events.PRECOMMIT_UPDATE:
            if self._should_regenerate_flowc(driver_context):
                for chain in self._get_chains_by_classifier_id(
                        driver_context._plugin_context,
                        driver_context.current['id']):
                    c_ctx = sfc_ctx.PortChainContext(
                        driver_context._plugin, driver_context._plugin_context,
                        chain, chain)
                    self.update_port_chain_precommit(c_ctx, remap=True)

    def _handle_port_bound(self, rtype, event, trigger, driver_context,
                           **kwargs):
        if event == events.PRECOMMIT_UPDATE:
            context = driver_context
            p_ctx = driver_context._plugin_context
            port_id = context.current['id']
            c_host = context.host
            o_host = context.original_host
            c_bound = self.aim_mech._is_port_bound(context.current)
            o_bound = self.aim_mech._is_port_bound(context.original)
            c_active = context.current.get(
                'status') == n_constants.PORT_STATUS_ACTIVE
            o_active = context.original.get(
                'status') == n_constants.PORT_STATUS_ACTIVE
            if c_bound and c_active and ((c_host != o_host) or
                                         (c_bound != o_bound) or
                                         (c_active != o_active)):
                LOG.debug("Update port pair for port %s", context.current)
                pps = self.sfc_plugin.get_port_pairs(
                    p_ctx, filters={'ingress': [driver_context.current['id']]})
                pps.extend(self.sfc_plugin.get_port_pairs(
                    p_ctx, filters={'egress': [driver_context.current['id']]}))
                for pp in pps:
                    # Only update if both ports are bound
                    other = pp['ingress'] if pp['ingress'] != port_id else (
                        pp['egress'])
                    other = self.plugin.get_port(p_ctx, other)
                    other_bound = self.aim_mech._is_port_bound(other)
                    other_active = other.get(
                        'status') == n_constants.PORT_STATUS_ACTIVE
                    if other_bound and other_active:
                        d_ctx = sfc_ctx.PortPairContext(self.sfc_plugin, p_ctx,
                                                        pp, pp)
                        self.update_port_pair_precommit(d_ctx, remap=True)

    def _handle_net_gbp_change(self, rtype, event, trigger, context,
                               network_id, **kwargs):
        chains = {}
        ppg_ids = self._get_group_ids_by_network_ids(context, [network_id])
        flowc_ids = self.aim_flowc._get_classifiers_by_network_id(
            context, network_id)
        for flowc_id in flowc_ids:
            for chain in self._get_chains_by_classifier_id(context,
                                                           flowc_id):
                chains[chain['id']] = chain
        if rtype == constants.GBP_NETWORK_VRF:
            # Don't need to check PPGs if the EPG is changing
            for ppg_id in ppg_ids:
                for chain in self._get_chains_by_ppg_ids(context, [ppg_id]):
                    chains[chain['id']] = chain
        for chain in chains.values():
            flowcs, ppgs = self._get_pc_flowcs_and_ppgs(context, chain)
            self._validate_port_chain(context, chain, flowcs, ppgs)

    def _handle_net_link_change(self, rtype, event, trigger, context,
                                networks_map, host_links, host, **kwargs):
        aim_ctx = aim_context.AimContext(db_session=context.session)
        cdis = self.aim.find(aim_ctx, aim_sg.ConcreteDeviceInterface,
                             host=host)
        cdi_by_port = {}
        for cdi in cdis:
            cdi_by_port.setdefault(
                self.name_mapper.reverse_port(context.session, cdi.name),
                []).append(cdi)
        ports = self.plugin.get_ports(context,
                                      filters={'id': cdi_by_port.keys()})
        networks_map = {x[0]['id']: x[1] for x in networks_map}

        for port in ports:
            if port['network_id'] not in networks_map:
                LOG.warning("We found a ConcreteDeviceInterface for port %s "
                            "and host %s, but no segment associated to it." %
                            (port['id'], host))
                continue
            hlinks = self.aim_mech._filter_host_links_by_segment(
                context.session, networks_map[port['network_id']], host_links)
            path = '' if not hlinks else hlinks[0].path
            for cdi in cdi_by_port.get(port['id']):
                self.aim.update(aim_ctx, cdi, path=path)

    def _get_flowc_src_network(self, plugin_context, flowc):
        return self.plugin.get_network(
            plugin_context, flowc['l7_parameters'][sfc_cts.LOGICAL_SRC_NET])

    def _get_flowc_dst_network(self, plugin_context, flowc):
        return self.plugin.get_network(
            plugin_context, flowc['l7_parameters'][sfc_cts.LOGICAL_DST_NET])

    def _get_pc_flowcs_and_ppgs(self, plugin_context, pc):
        flowcs = self.flowc_plugin.get_flow_classifiers(
            plugin_context, filters={'id': pc['flow_classifiers']})
        unordered_ppgs = self.sfc_plugin.get_port_pair_groups(
            plugin_context, filters={'id': pc['port_pair_groups']})
        # Keep order
        ppgs = []
        for ppg_id in pc['port_pair_groups']:
            for ppg in unordered_ppgs:
                if ppg['id'] == ppg_id:
                    ppgs.append(ppg)
                    break
        return flowcs, ppgs

    def _get_flowc_network_group(self, plugin_context, flowc, net, prefix):
        flc_aid = self._get_external_group_aim_name(plugin_context, flowc,
                                                    prefix)
        l3out = self.aim_mech._get_svi_net_l3out(net)
        if l3out:
            # Create ExternalNetwork and ExternalSubnet on the proper
            # L3Out. Return the External network
            ext_net = aim_resource.ExternalNetwork(
                tenant_name=l3out.tenant_name, l3out_name=l3out.name,
                name=flc_aid)
            return ext_net
        else:
            return self.aim_mech._get_epg_by_network_id(plugin_context.session,
                                                        net['id'])

    def _get_flowc_provider_group(self, plugin_context, flowc):
        net = self._get_flowc_dst_network(plugin_context, flowc)
        return self._get_flowc_network_group(plugin_context, flowc, net,
                                             FLOWC_DST)

    def _get_flowc_consumer_group(self, plugin_context, flowc):
        net = self._get_flowc_src_network(plugin_context, flowc)
        return self._get_flowc_network_group(plugin_context, flowc, net,
                                             FLOWC_SRC)

    def _map_flowc_provider_group(self, plugin_context, flowc):
        aim_ctx = aim_context.AimContext(plugin_context.session)
        net = self._get_flowc_dst_network(plugin_context, flowc)
        return self.aim.get(aim_ctx, self._map_flowc_network_group(
            plugin_context, net, flowc['destination_ip_prefix'], flowc,
            FLOWC_DST))

    def _map_flowc_consumer_group(self, plugin_context, flowc):
        aim_ctx = aim_context.AimContext(plugin_context.session)
        net = self._get_flowc_src_network(plugin_context, flowc)
        return self.aim.get(aim_ctx, self._map_flowc_network_group(
            plugin_context, net, flowc['source_ip_prefix'], flowc, FLOWC_SRC))

    def _get_flowc_provider_vrf(self, plugin_context, flowc):
        net = self._get_flowc_dst_network(plugin_context, flowc)
        return self.aim_mech._get_vrf_by_network(plugin_context.session, net)

    def _get_flowc_consumer_vrf(self, plugin_context, flowc):
        net = self._get_flowc_src_network(plugin_context, flowc)
        return self.aim_mech._get_vrf_by_network(plugin_context.session, net)

    def _get_external_group_aim_name(self, plugin_context, flowc, prefix):
        if prefix == FLOWC_SRC:
            cidr = flowc['source_ip_prefix']
            net = self._get_flowc_src_network(plugin_context, flowc)
        else:
            cidr = flowc['destination_ip_prefix']
            net = self._get_flowc_dst_network(plugin_context, flowc)
        cidr = netaddr.IPNetwork(cidr)
        if cidr.prefixlen == 0:
            cidr = 'default'
        else:
            cidr = aim_utils.sanitize_display_name(str(cidr))
        name = self.name_mapper.network(plugin_context.session, net['id'],
                                        prefix=cidr + '_')
        return name

    def validate_aim_mapping(self, mgr):
        # REVISIT: Register all AIM resource types used by the SFC
        # mapping but not the Neutron or GBP mappings.

        # REVISIT:  Register DB tables to be validated.

        # Determine expected AIM resources and DB records for each
        # SFC resource type.
        self._validate_flow_classifiers(mgr)
        self._validate_port_pair_groups(mgr)
        self._validate_port_chains(mgr)

    def _validate_flow_classifiers(self, mgr):
        # REVISIT: Implement validation of actual mapping to AIM
        # resources.
        if mgr.actual_session.query(flowc_db.FlowClassifier).first():
            mgr.validation_failed(
                "SFC->AIM validation for FC not yet implemented")

    def _validate_port_pair_groups(self, mgr):
        # REVISIT: Implement validation of actual mapping to AIM
        # resources.
        if mgr.actual_session.query(sfc_db.PortPairGroup).first():
            mgr.validation_failed(
                "SFC->AIM validation for PPG not yet implemented")

    def _validate_port_chains(self, mgr):
        # REVISIT: Implement validation of actual mapping to AIM
        # resources.
        if mgr.actual_session.query(sfc_db.PortChain).first():
            mgr.validation_failed(
                "SFC->AIM validation for PC not yet implemented")
